import os
import sys
forge = sys.modules["forge"]

# NOTE: The Maya version of boost may be different.
# The boost-regex appears to be sensitive and they don't provide the dso.
# Moving aside their include/boost seems to fix regex,
# but it leaves their other boost dso's cross-versioned with system includes.

def setup(module):
	srcList = [	"maya.pmh",
				"MayaBrush",
				"MayaDraw",
				"MayaGraph",
				"MayaNode",
				"SurfaceAccessibleMaya",
				"SurfaceAccessorMaya",
				"SurfaceTrianglesMaya",
				"assertMaya",
				"mayaDL" ]

	# force all of FE to use Maya's version of boost
#	forge.includemap['maya'] = os.path.join(os.getenv('HT'),'include')

	dll = module.DLL( "fexMayaDL", srcList )

	# this overrides previous -W flags in codegen
	forge.cppmap['codegen'] += ' -Wno-deprecated -Wall -W -Wno-parentheses -Wno-sign-compare -Wno-reorder -Wno-uninitialized -Wunused -Wno-unused-parameter'

	deplibs = forge.corelibs+ [
				"fexDrawDLLib",
				"fexMetaDLLib",
				"fexSurfaceDLLib",
				"fexWindowLib"	]

	maya_path = os.getenv('MAYA_ROOT')

	dll.linkmap = {}
	if forge.fe_os == "FE_LINUX":
		dll.linkmap["mayalibs"] = "-shared"
		dll.linkmap['mayalibs'] += " -L"+os.path.join(maya_path,'lib')
		dll.linkmap["mayalibs"] += " -lOpenMaya -lOpenMayaUI"
		dll.linkmap["mayalibs"] += " -Xlinker -rpath='$ORIGIN/../'"
	elif forge.fe_os == "FE_WIN32" or forge.fe_os == "FE_WIN64":
		dll.linkmap["mayalibs"] = "TODO"

	forge.deps( ["fexMayaDLLib"], deplibs )

	if maya_path:
		module.includemap = { 'maya' : os.path.join(maya_path,'include') }

	# TODO should work on non-linux
	mayaLib = 'libfexMayaDL.so'
	mayaLibPath = os.path.join(forge.libPath, 'maya')
	if os.path.isdir(mayaLibPath) == 0:
		os.mkdir(mayaLibPath, 0755)
	source = os.path.join('..', mayaLib)
	dest = os.path.join(mayaLibPath, mayaLib)
	if os.path.lexists(dest) == 0:
		os.symlink(source, dest)

	fe_maya_libs = 'fexMayaDL'
	if os.path.exists('ext/grass'):
		fe_maya_libs += ':fexGrassDL'
		sys.stdout.write("grass ")
	if os.path.exists('ext/vegetation'):
		fe_maya_libs += ':fexVegetationDL'
		sys.stdout.write("vegetation ")
	module.cppmap = {}
	module.cppmap['maya'] = ' -DFE_MAYA_LIBS=' + fe_maya_libs
	module.Module('test')

def auto(module):
	test_file = """
#define _BOOL
#include <maya/MTypes.h>

int main(void)
{
	return(0);
}
    \n"""

	maya_path = os.getenv('MAYA_ROOT')
	if maya_path:
		forge.includemap['maya'] = os.path.join(maya_path,'include')

	return forge.cctest(test_file)
